#######################################################################
# Copyright (c) 2025 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class ThreatIntelligencesController < ApplicationController
  self.view_paths = 'lib/app/views'
  layout "#{Rails.root}/app/views/layouts/application"
  prepend_view_path "#{Rails.root}/app/views"

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================

  before_action :set_reputation_and_check_privileges,
                only: [:edit, :update, :destroy]
  before_action :set_rep_policy_and_check_privileges,
                only: [:new, :create, :update, :reorder]
  ## ======================================================================
  # => Actions
  ## ======================================================================

  def new
    @sensor = Sensor.find(params[:sensor_id]) if params[:sensor_id]
    @threat_intelligence = @ti_policy.threat_intelligences
                  .build(ti_type: params[:ti_type])
    ti_type_str = ThreatIntelligence::TI_TYPE_MAP[params[:ti_type].to_i]
    @threat_sources = ThreatIntelSource.where(threat_src_type: ti_type_str)
    
    render layout: false
  end

  def create
    @threat_intelligence = @ti_policy.threat_intelligences.build(threat_intelligence_params)
  
    respond_to do |format|
      if @threat_intelligence.save
        type_string = ThreatIntelligence::TI_TYPE_MAP[@threat_intelligence.ti_type]
        @threat_sources = ThreatIntelSource.where(threat_src_type: type_string)
                                           .pluck(:id, :name)
                                           .map { |id, name| [name, id] }

        format.html { redirect_to some_path }
        format.js
        format.json { head :no_content }
      else
        type_string = ThreatIntelligence::TI_TYPE_MAP[@threat_intelligence.ti_type]
        @threat_sources = ThreatIntelSource.where(threat_src_type: type_string)
  
        format.html { render :new, layout: false, status: :unprocessable_entity }
        format.js
        format.json { render json: @threat_intelligence.errors, status: :unprocessable_entity }
      end
    end
  end

  def edit
    @threat_intelligence = ThreatIntelligence.find(params[:id])

    ti_type_str = ThreatIntelligence::TI_TYPE_MAP[@threat_intelligence.ti_type]
    @threat_sources = ThreatIntelSource.where(threat_src_type: ti_type_str)

    render layout: false
  end

  def update
    @threat_intelligence = ThreatIntelligence.find(params[:id])
  
    respond_to do |format|
      if @threat_intelligence.update(threat_intelligence_params)
        type_string = ThreatIntelligence::TI_TYPE_MAP[@threat_intelligence.ti_type]
        @threat_sources = ThreatIntelSource.where(threat_src_type: type_string)
                                           .pluck(:id, :name)
                                           .map { |id, name| [name, id] }
                                           
        format.html { redirect_to some_path }
        format.js
        format.json { head :no_content }
      else
        type_string = ThreatIntelligence::TI_TYPE_MAP[@threat_intelligence.ti_type]
        @threat_sources = ThreatIntelSource.where(threat_src_type: type_string)

        format.html { render :edit, layout: false, status: :unprocessable_entity }
        format.js
        format.json { render json: @threat_intelligence.errors, status: :unprocessable_entity }
      end
    end
  end

  def destroy
    delete_memcached_key(@threat_intelligence)
  
    @threat_intelligence.destroy
  
    if params[:sensor_id]
      sensor = Sensor.find(params[:sensor_id])
      redirect_to sensor_ti_policy_path(sensor, @threat_intelligence.ti_policy)
    else
      redirect_to ti_policy_path(@threat_intelligence.ti_policy)
    end
  end

  def reorder
    ti_type = params[:ti_type]
    order = params[:positions].split(';')
    order.each_with_index do |auto_id, position|
      ThreatIntelligence.find_by(id: auto_id,
                         ti_type: ti_type,
                         ti_policy_id: @ti_policy.id)
        .update(position: (position + 1))
    end

    render js: ''
  end

  private

  def delete_memcached_key(threat_intel)
    return if threat_intel.ti_type.nil? || threat_intel.threat_intel_source_id.blank?

    memcached_namespace = 'rbti'
    policy_id = threat_intel.threat_intel_source_id
    action_str = threat_intel.category == 1 ? 'c' : 'm'
    type_str = ThreatIntelligence::TI_TYPE_MAP[threat_intel.ti_type] || threat_intel.ti_type.to_s

    source = ThreatIntelSource.find_by(id: threat_intel.threat_intel_source_id)
    return unless source

    values = fetch_values_from_threat_source(source, threat_intel.ti_type)
    return if values.empty?

    keys = values.map do |val|
      "#{memcached_namespace}:#{policy_id}:#{type_str}:#{action_str}:#{val.downcase}"
    end

    MemcachedKeyInvalidator.new.invalidate_keys(keys)
  end

  def fetch_values_from_threat_source(source, ti_type)
    return [] unless source&.enable

    cache_file = Rails.root.join("tmp", "ti_#{source.id}.cache").to_s
    return [] unless File.exist?(cache_file)

    raw_lines = File.readlines(cache_file).map(&:strip).reject { |line| line.blank? || line.start_with?('#') }

    case ti_type
    when ThreatIntelligence::IP_TYPE
      raw_lines.select { |line| IPAddr.new(line) rescue false }.map(&:strip)
    when ThreatIntelligence::URL_TYPE
      raw_lines.select do |line|
        uri = URI.parse(line) rescue nil
        uri.is_a?(URI::HTTP) || uri.is_a?(URI::HTTPS)
      end
    when ThreatIntelligence::DOMAIN_TYPE
      raw_lines.grep(/\A[a-z0-9.-]+\z/i)
    when ThreatIntelligence::SHA1_TYPE
      raw_lines.grep(/\A[a-f0-9]{40}\z/i)
    when ThreatIntelligence::SHA2_TYPE
      raw_lines.grep(/\A[a-f0-9]{64}\z/i)
    else
      []
    end
  end

  def set_rep_policy_and_check_privileges
    ti_policy_id = if params[:ti]
                             params[:ti][:ti_policy_id]
                           else
                             params[:ti_policy_id]
                           end

    @ti_policy = TiPolicy.find(ti_policy_id)
    return unless current_user.cannot? :manage, @ti_policy.owner
    respond_to do |format|
      format.html do
        redirect_to sensors_path,
                    alert: "#{I18n.t('controllers.no_privileges')}"
      end
      format.js do
        render json: { error: "#{I18n.t('controllers.no_privileges')}" },
               status: :unauthorized
      end
    end
  end

  def set_reputation_and_check_privileges
    @threat_intelligence = ThreatIntelligence.find(params[:id])
    return unless current_user.cannot? :manage, @threat_intelligence.ti_policy.owner
    respond_to do |format|
      format.html do
        redirect_to sensors_path,
                    alert: "#{I18n.t('controllers.no_privileges')}"
      end
      format.js do
        render json: { error: "#{I18n.t('controllers.no_privileges')}" },
               status: :unauthorized
      end
    end
  end

  def threat_intelligence_params
    params.require(:threat_intelligence)
      .permit(:ti_policy_id, :category, :value, :ti_type, :threat_intel_source_id, :weight)
  end
end
