#######################################################################
# Copyright (c) 2025 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class ThreatIntelligence < ApplicationRecord
  include Redborder::Audit

  IP_TYPE = 1
  URL_TYPE = 2
  DOMAIN_TYPE = 3
  SHA1_TYPE = 4
  SHA2_TYPE = 5

  CATEGORY = {
    1 => { name: 'clean', color: 'green', category_2: 'allow' },
    2 => { name: 'malicious', color: 'red', category_2: 'deny' }
  }

  TI_TYPE_MAP = {
    IP_TYPE => 'ip',
    URL_TYPE => 'url',
    DOMAIN_TYPE => 'domain',
    SHA1_TYPE => 'sha1',
    SHA2_TYPE => 'sha2'
  }
  
  before_validation :set_weight_for_clean_category

  belongs_to :ti_policy
  belongs_to :threat_intel_source, required: true

  validates :ti_type, :category, numericality: { only_integer: true }

  validate :validate_threat_list_constraints

  after_save :update_sensors
  before_destroy :update_sensors
  before_create :set_position

  audited on: [:create, :update, :destroy]

  def humanize_value
    threat_intel_source.try(:name)
  end

  def category_name
    CATEGORY[category][:name]
  end

  def category_color
    CATEGORY[category][:color]
  end

  def value=(v)
    begin
      super(NetAddr::CIDR.create(v).to_s.gsub(/\/32$/, ""))
    rescue
      super(v)
    end
  end

  def ti_type_name
    TI_TYPE_MAP[ti_type].upcase
  end

  private

  def update_sensors
    Sensor.where(ti_policy: self.ti_policy).update_all(need_ti_apply: true)
  end

  def set_position
    return if position?
    self.position = 1 + (ti_policy.threat_intelligences.where(ti_type: ti_type).maximum(:position) || 0)
  end

  def set_weight_for_clean_category
    if category == 1
      self.weight = 0.0
    end
  end

  def validate_threat_list_constraints
    validate_weight
    validate_unique_threat_intel_source
  end

  def validate_weight
    return if category == 1
    unless weight.present? && weight > 0 && weight <= 1
      errors.add(:weight, "Must be between 0.01 and 1")
    end
  end

  def validate_unique_threat_intel_source
    existing_record = ThreatIntelligence.where(
      ti_policy_id: ti_policy_id,
      threat_intel_source_id: threat_intel_source_id
    ).where.not(id: id).first
  
    if existing_record
      errors.add(:threat_intel_source_id, "‘#{threat_intel_source.try(:name)}’ It’s already on the list.")
    end
  end

end
