#######################################################################
# Copyright (c) 2025 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

#
# Define threat intelligence policies
#
class TiPolicy < ApplicationRecord
  # Includes
  include Redborder::Audit

  # -------------------------------------------------------------
  # Relations
  # -------------------------------------------------------------
  has_many :sensors
  has_many :threat_intelligences, -> { order('position ASC') }
  belongs_to :owner, class_name: "Sensor", foreign_key: "sensor_id"

  before_validation :set_default_thresholds
  has_many :threat_intelligences, dependent: :destroy

  # -------------------------------------------------------------
  # Validations
  # -------------------------------------------------------------
  validates :threshold_ip, :threshold_url, :threshold_domain, :threshold_sha1, :threshold_sha2,
            numericality: { only_integer: true, greater_than_or_equal_to: 0, less_than_or_equal_to: 100 }

  # -------------------------------------------------------------
  # Audited
  # -------------------------------------------------------------
  audited only: [:name, :description], on: [:create, :update, :destroy]

  # -------------------------------------------------------------
  # Methods
  # -------------------------------------------------------------

  #
  # Clean the name based changing spaces and . with underscores
  #
  def clean_name
    name.gsub(/\s+|\.+/, '_')
  end

  #
  # Return rules as CSV format
  def rules_to_csv_format
    header = %w(Rule_Type_Id Category_Id Value)
    csv = header.join(',') + "\n"

    # Iterate over the threat intelligence sorted by threat_intelligence_type and position
    threat_intelligence.order(:threat_intelligence_type, :position).each do |reputation|
      # Ignore rows related to threat_list_reputations
      next if reputation.threat_intelligence_type == Ti::THREAT_LIST_TYPE

      # Add row to CSV in the desired format
      csv += [reputation.threat_intelligence_type, reputation.action, reputation.value].join(',') + "\n"
    end

    csv
  end

  #
  # Clones current threat intelligence policy on the specified sensor with a new name
  #
  # == Parameters:
  # sensor_id::
  #   Integer with the ID of the sensor to clone the reputaiton policies
  #
  # == Returns:
  # Threat Intelligence policy object or nil
  #
  def clone(to_sensor_id)
    matched = /(.*) \((\d+)\)$/.match(name)
    current_name = matched.nil? ? name : matched[1]
    current_number = matched.nil? ? 0 : matched[2].to_i
  
    loop do
      current_number += 1
      clone_name = "#{current_name} (#{current_number})"
      next if TiPolicy.find_by(name: clone_name)
  
      clone_attributes = {
        name: clone_name,
        description: description,
        sensor_id: to_sensor_id
      }
  
      %w[ip url domain sha1 sha2].each do |type|
        threshold_field = "threshold_#{type}"
        clone_attributes[threshold_field.to_sym] = self.send(threshold_field)
      end
  
      created = TiPolicy.new(clone_attributes)
  
      if created.save
        # Clonar threat_intelligences completamente
        threat_intelligences.each do |ti|
          created.threat_intelligences.create(
            ti_type: ti.ti_type,
            value: ti.value,
            category: ti.category,
            position: ti.position,
            weight: ti.weight,
            threat_intel_source_id: ti.threat_intel_source_id
          )
        end
        return created
      end
    end
  end  

  # -------------------------------------------------------------
  # Work with files
  # -------------------------------------------------------------

  #
  # Add rules passed from csv file into reputation policy. If the rule is
  # present in reputation policy, the action will be updated.
  #
  def add_rules_from_file(content_file, category=nil)
    file = File.open(content_file.tempfile)
    updated_count = 0
    created_count = 0

    Ti.transaction do
      file.each_with_index do |line, index|
        next if index.zero?

        columns = line.sub(/[\r\n]+|[\n]+|[\r]+/, '').split(',')

        # Check how many columns is the csv file.
        case columns.size
        when 1
          value = columns[0]
          reputation_type = (value.length == 2 ? Ti::COUNTRY_TYPE : Ti::IP_NET_TYPE)
        when 2
          action, value = columns
          reputation_type = (value.length == 2 ? Ti::COUNTRY_TYPE : Ti::IP_NET_TYPE)
        else
          reputation_type, action, value = columns
        end

        flow_reputation = threat_intelligence.find_by(reputation_type: reputation_type, value: value)

        if flow_reputation and action.to_i != flow_reputation.action
          flow_reputation.update(action: action)
          updated_count += 1 if flow_reputation.errors.blank?
        elsif flow_reputation.nil?
          flow_reputation = Ti.create(flow_reputation_policy_id: self.id, action: action, reputation_type: reputation_type, value: value)
          created_count += 1 if flow_reputation.errors.blank?
        end
      end
    end
    file.close
    return created_count, updated_count
  end

  #
  # Add rules passed from csv file into reputation policy.
  # The previous rules will be deleted.
  #
  def replace_rules_from_file(content_file, action=nil)
    file = File.open(content_file.tempfile)
    created_count = 0

    Ti.transaction do
      threat_intelligence.delete_all

      file.each_with_index do |line, index|
        next if index.zero?

        columns = line.sub(/[\r\n]+|[\n]+|[\r]+/, '').split(',')

        # Check how many columns is the csv file.
        case columns.size
        when 1
          value = columns[0]
          reputation_type = (value.length == 2 ? Ti::COUNTRY_TYPE : Ti::IP_NET_TYPE)
        when 2
          action, value = columns
          reputation_type = (value.length == 2 ? Ti::COUNTRY_TYPE : Ti::IP_NET_TYPE)
        else
          reputation_type, action, value = columns
        end

        flow_reputation = Ti.create(flow_reputation_policy_id: self.id, action: action, reputation_type: reputation_type, value: value)
        created_count += 1 if reputation.errors.blank?
      end
    end
    file.close
    created_count
  end

  # Método para obtener el threshold según el tipo
  def threshold_for_type(ti_type)
    case ti_type.to_s
    when 'ip' then threshold_ip
    when 'url' then threshold_url
    when 'domain' then threshold_domain
    when 'sha1' then threshold_sha1
    when 'sha2' then threshold_sha2
    else 0
    end
  end
  
  # Método para establecer threshold según el tipo
  def set_threshold_for_type(ti_type, value)
    case ti_type.to_s
    when 'ip' then self.threshold_ip = value
    when 'url' then self.threshold_url = value
    when 'domain' then self.threshold_domain = value
    when 'sha1' then self.threshold_sha1 = value
    when 'sha2' then self.threshold_sha2 = value
    end
  end

  private

  def set_default_thresholds
    self.threshold_ip = 0 if threshold_ip.nil?
    self.threshold_url = 0 if threshold_url.nil?
    self.threshold_domain = 0 if threshold_domain.nil?
    self.threshold_sha1 = 0 if threshold_sha1.nil?
    self.threshold_sha2 = 0 if threshold_sha2.nil?
  end
  
end