#######################################################################
# Copyright (c) 2025 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

# Class responsible for invalidating keys on a Memcached client (using Dalli)
class MemcachedKeyInvalidator
    # Initializes the class with an instance of Dalli::Client
    # @param dalli_client [Dalli::Client] an instance of the Memcached client
    def initialize
        config = YAML.load_file(Rails.root.join('config', 'memcached_config.yml'))[Rails.env].symbolize_keys rescue {}
        @dalli_client = Dalli::Client.new(
          config[:servers],
          (config[:options] || {}).merge(namespace: 'rbti', compress: true)
        )
    end
  
    # Invalidates a key in Memcached by reducing its time to live (TTL) to 10 seconds
    #
    # @param key [String] the key to invalidate
    # @return [Boolean] false if the key does not exist, true if it was updated successfully
    def invalidate(key)
      # Get the current value of the key
      value = @dalli_client.get(key)
      
      # If it does not exist, it cannot be invalidated
      return false unless value
  
      # Resave the value with a shorter TTL (10 seconds)
      # This causes the value to "expire" early, acting as a soft invalidation
      @dalli_client.set(key, value, 10)
    end
  
    # Invalidate multiple keys by iterating over them
    #
    # @param keys [Array<String>] list of keys to invalidate
    def invalidate_keys(keys)
      # Iterate over each key and apply invalidation individually
      keys.each { |key| invalidate(key) }
    end
end
