#######################################################################
# Copyright (c) 2025 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Jobs
    class UpdateThreatIntelJob < Struct.new(:verbose, :user_id)
      include Redborder::Jobs::JobHelper

      MEMCACHED_CONFIG = (YAML.load_file('config/memcached_config.yml')[Rails.env].symbolize_keys rescue {}).freeze
      MEMCACHED_NAMESPACE = 'rbti'
      TMP_DIR = Rails.root.join('tmp')
      CACHE_DEFAULT_VALUE = 1
      CACHE_EXPIRATION = ((Setting.find_by(name: 'threat_intel_sources_download_periodicity')&.value.to_i || 10) + 2).minutes

      def perform
        content = +""

        begin
          Delayed::Worker.logger.info "Starting UpdateThreatIntelJob with user_id: #{user_id}"
          Delayed::Worker.logger.info "Verbose mode: #{verbose}"
          Delayed::Worker.logger.info "Memcached servers: #{MEMCACHED_CONFIG[:servers].join(', ')}"
          Delayed::Worker.logger.info "Memcached namespace: #{MEMCACHED_NAMESPACE}"
          content << "Starting UpdateThreatIntelJob with user_id: #{user_id}\n"
          content << "Verbose mode: #{verbose}\n"
          content << "Memcached servers: #{MEMCACHED_CONFIG[:servers].join(', ')}\n"
          content << "Memcached namespace: #{MEMCACHED_NAMESPACE}\n"

          content << "\nTHREAT INTELLIGENCE SOURCES\n\n"
          Delayed::Worker.logger.info "Downloading threat intelligence sources..."
          download_threat_intel_sources(content)
          
          Delayed::Worker.logger.info "Completed processing threat intelligence sources."
          content << "\nTHREAT INTELLIGENCE POLICY\n\n"
          process_sensor_policies(content)
        rescue => e
          log_exception(content, e)
        ensure
          content << reschedule_if_needed
          if content.match?(/(ERROR|Exception)/i)
            content << "\nJob completed with errors.\n"
          else
            content << "\nJob completed successfully.\n"
          end        
          StoredDelayedJob.create(job: self.class.name.demodulize, user_id: user_id, content: content)
        end
      end

      private

      # Processes configured threat intelligence feeds
      def download_threat_intel_sources(content)
        sources = ThreatIntelSource.all
        return if sources.empty?

        enabled, disabled = sources.partition(&:enable)

        enabled.each_with_index do |source, i|
          log_source_header(content, source, i + 1, enabled.size)
          client = init_client(source, content)
          next unless client

          begin
            cache_file = TMP_DIR.join("ti_#{source.id}.cache").to_s
            client.download(remote_path: source.path, local_path: cache_file, type: source.threat_src_type)
            values = filter_values_by_type(cache_file, source.threat_src_type)

            content << "Downloaded and cached to #{File.join('/tmp', File.basename(cache_file))}\n"
            content << "Parsed values: #{values.size}\n"
            content << "WARNING: No valid values found - file NOT created.\n" if values.empty?
          rescue => e
            log_exception(content, e, "Download error for #{source.path}")
          end

          content << "\n --------------------------------------------\ \n"
        end

        disabled.each do |source|
          content << "Entity #{source.id} (#{source.path}) is disabled, skipping download.\n\n --------------------------------------------\ \n"
        end
      end

      # Applies threat intelligence policies to sensors
      def process_sensor_policies(content)
        content << "\nTHREAT INTELLIGENCE POLICY\n\n--------------------------------------------\n"
        ti_policy_found = false
        
        # Skip sensor type IPS_GROUP (3) IPS_BINDING(4)
        Sensor.includes(ti_policy: :threat_intelligences).where.not(type: [3, 4]).find_each do |sensor|
          node_policy = sensor.chef_node&.dig('redborder', 'ti_policy_id')
          next if sensor.ti_policy.blank? && node_policy.blank?

          ti_policy_found = true

          if sensor.ti_policy_id.blank?
            content << "Sensor #{sensor.name} has no Threat Intelligence Policy assigned. Skipping...\n"
          elsif node_policy != sensor.ti_policy_id
            content << "Configuration mismatch for sensor #{sensor.name}. Expected: #{sensor.ti_policy_id}, Node Policy: #{node_policy}.\n"
          else
            content << "\nSensor: #{sensor.name}\n"
            content << "Applying Threat Intelligence Policy\n"
            content << apply_threat_intel_policy(sensor)
          end
        end

        content << "No sensors with threat intelligence policies.\n" unless ti_policy_found
      end

      # Initializes the client to download the remote font
      def init_client(entity, content)
        data_bag = Redborder::Databag.load('passwords', 'threat_intel_sources') rescue {}
        password = entity.anonymous ? "" : (data_bag[entity.id.to_s] || "")

        Redborder::ThreatIntel::Client.new(
          protocol: ThreatConnType.find_by(id: entity.threat_conn_types_id)&.name,
          host: entity.host,
          port: entity.port,
          user: entity.anonymous ? "anonymous" : entity.user,
          password: password,
          dalli_options: {
            servers: MEMCACHED_CONFIG[:servers],
            options: { namespace: MEMCACHED_NAMESPACE, compress: true }
          }
        )
      rescue => e
        content << "Error initializing client: #{e.class}: #{e.message}\n"
        nil
      end

      # Apply intelligence policy to a sensor
      def apply_threat_intel_policy(sensor)
        content = +""
        sensor.update(applying: false)
        options = MEMCACHED_CONFIG[:options] || {}

        # Memcached Client
        dalli_client = Dalli::Client.new(
          MEMCACHED_CONFIG[:servers],
          options.merge(namespace: MEMCACHED_NAMESPACE, compress: true)
        )
        policy = sensor.ti_policy
        totals = Hash.new { |h, k| h[k] = { inserted: 0, deleted: 0 } }

        # Weight validation by type
        weights_by_type = policy.threat_intelligences.select { |e| e.category == 2 }.group_by(&:ti_type).transform_values { |v| v.sum(&:weight) }
        invalid_weights = weights_by_type.select { |_, w| w.round(2) != 1.0 }

        unless invalid_weights.empty?
          content << "\t[WARN] Threat list weights must sum exactly to 1.0 for each type.\n"
          invalid_weights.each do |type, w|
            type_str = (ThreatIntelligence::TI_TYPE_MAP[type] || type.to_s).upcase
            content << "\t[WARN] - #{type_str}: Current weight: #{w}\n"
          end          
          content << "\tThreat lists will NOT be applied.\n"
          return content
        end

        # We group and apply by source
        policy.threat_intelligences.group_by(&:threat_intel_source_id).each do |source_id, entries|
          cache_file = TMP_DIR.join("ti_#{source_id}.cache")
          source = ThreatIntelSource.find_by(id: source_id)

          if File.exist?(cache_file)
            content << "Using cached #{source.threat_src_type.upcase} list for source #{source.name} (#{File.basename(cache_file)})\n"
          else
            content << "No cache found for source #{source.name}, skipping...\n"
            next
          end
          
          values = File.readlines(cache_file).map(&:strip).reject(&:empty?)
          filtered = filter_lines(values, source.threat_src_type)

          entries.each do |entry|
            ti_type_str = ThreatIntelligence::TI_TYPE_MAP[entry.ti_type] || "unknown"
            category = entry.category == 1 ? 'c' : 'm'
            weight = entry.category == 2 ? entry.weight.to_f : CACHE_DEFAULT_VALUE

            filtered.each do |val|
              key = "#{sensor.ti_policy_id}:#{ti_type_str}:#{category}:#{val.downcase}"
              clean_opposite_key_if_needed(dalli_client, val, entry, totals)
              if set_key(dalli_client, key, weight)
                totals[entry.ti_type][:inserted] += 1 
                Delayed::Worker.logger.info "Inserted key: #{key} with weight: #{weight}"
              end
            end
          end
        end

        content << "Threat Intelligence Policy name: #{policy.name}\nThreat Intelligence types:\n"
        if totals.values.all? { |v| v[:inserted] == 0 && v[:deleted] == 0 }
          content << "\tNo changes applied.\n"
        else
          totals.each do |type, c|
            type_str = (ThreatIntelligence::TI_TYPE_MAP[type] || type.to_s).upcase
            content << "\t - #{type_str}: #{c[:inserted]} inserted, #{c[:deleted]} deleted.\n"
          end
        end

        content
      end

      # Remove opposite keys if they are replaced by the new category
      def clean_opposite_key_if_needed(client, value, entry, totals)
        cat = entry.category == 1 ? 2 : 1
        opp_key = "#{entry.ti_policy_id}:#{ThreatIntelligence::TI_TYPE_MAP[entry.ti_type]}:#{cat == 1 ? 'c' : 'm'}:#{value.downcase}"

        opp_entries = ThreatIntelligence.where(
          ti_policy_id: entry.ti_policy_id,
          ti_type: entry.ti_type,
          value: value,
          category: cat
        )

        if opp_entries.any? && opp_entries.max_by(&:position).position.to_i < entry.position.to_i
          if client.delete(opp_key)
            totals[entry.ti_type][:deleted] += 1
            Delayed::Worker.logger.info "Deleting opposite key: #{opp_key}"
          end
        end
      end

      # Insert value into Memcached with expiration
      def set_key(client, key, value)
        client.set(key, value, CACHE_EXPIRATION)
      end

      # Reschedule the job for the next execution
      def reschedule_if_needed
        return "\nNo need to reschedule this job as there is already one job pending in the queue.\n" unless DelayedJob.where('handler like ?', '%!ruby/struct:Redborder::Jobs::UpdateThreatIntelJob%').count <= 1

        interval = Setting.threat_intel_sources_download_periodicity.to_i || 10
        Delayed::Job.enqueue(self.class.new(verbose, user_id), run_at: Time.now + interval.minutes, priority: 10)
        return "\nJob rescheduled to run in #{interval} minutes.\n"
      end

      # Log exceptions to content
      def log_exception(content, exception, message = nil)
        content << "#{message}\n" if message
        content << "#{exception.class}: #{exception.message}\n"
        exception.backtrace.each { |line| content << "#{line}\n" }
      end

      # Initial log of each source
      def log_source_header(content, source, index, total)
        content << "[#{index}/#{total}] Downloading from #{source.name} (#{File.basename(source.path)})...\n"
        content << "Starting download for #{source.path}...\n"
        content << "Initializing client for entity #{source.id} (#{source.path})...\n"
      end

      # Filter valid values based on type
      def filter_values_by_type(file_path, type)
        return [] unless File.exist?(file_path)
        lines = File.readlines(file_path).map(&:strip).reject { |l| l.blank? || l.start_with?('#') }
        filter_lines(lines, type)
      end

      # Filter specific lines by type (IP, domain, URL, hashes)
      def filter_lines(lines, type)
        lines = lines.map(&:strip).reject(&:empty?)
      
        case type
        when 'ip'
          lines.select { |line| valid_ip?(line) }
        when 'domain'
          lines.select { |line| valid_domain?(line) }
        when 'url'
          lines.select { |line| valid_url?(line) }
        when 'sha1'
          lines.select { |line| valid_sha?(line, 40) }
        when 'sha2'
          lines.select { |line| valid_sha?(line, 64) }
        else
          lines
        end
      end
      
      def valid_ip?(ip)
        return false if ip.include?('/')
        
        IPAddr.new(ip)
        true
      rescue IPAddr::InvalidAddressError
        false
      end
      
      def valid_domain?(domain)
        domain.match?(/\A(?=.{1,253}\z)([a-z0-9](?:[a-z0-9\-]{0,61}[a-z0-9])?\.)+[a-z]{2,63}\z/i)
      end

      def valid_url?(url)
        uri = URI.parse(url)
        uri.is_a?(URI::HTTP) || uri.is_a?(URI::HTTPS)
      rescue URI::InvalidURIError
        false
      end
      
      def valid_sha?(hash, length)
        hash.match?(/\A[a-f0-9]{#{length}}\z/i)
      end

    end
  end
end
